///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//  Copyright  NetworkDLS 2002, All rights reserved
//
// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF 
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A 
// PARTICULAR PURPOSE.
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#ifndef _CXMLWriter_Cpp_
#define _CXMLWriter_Cpp_
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include <Windows.H>
#include <Stdio.H>
#include <Stdlib.H>

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

#include "../../@Libraries/Debug/Debug.H"
#include "../../@Classes/CMemPool/CMemPool.H"
#include "../../@Classes/CStringBuilder/CStringBuilder.H"
#include "../../@Libraries/NSWFL/NSWFL.H"

#include "CXMLReader.H"
#include "CXMLWriter.H"
#ifdef _CXML_USE_COMPRESSION
#include "CXMLCompression.H"
#endif

#ifdef _USE_GLOBAL_MEMPOOL
#include "../CMemPool/CMemPool.H"
extern CMemPool *pMem; //pMem must be defined and initalized elsewhere.
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CXMLWriter::CXMLWriter(const char *sChildName)
{
	this->Initialize(sChildName);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CXMLWriter::~CXMLWriter(void)
{
	this->Destroy();	
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CXMLWriter::CXMLWriter(void)
{
	this->pBuilder = NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CXMLReader *CXMLWriter::ToReader(CXMLReader *lpXMLReader)
{
	int iLength = this->Length();

	char *sText = (char *)calloc(iLength + 1, sizeof(char));
	if(sText)
	{
		iLength = this->Text(sText, iLength);
		if(lpXMLReader->Initialize(sText, iLength))
		{
			free(sText);
			return lpXMLReader;
		}
		free(sText);
	}
	return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CXMLWriter::GetName(char *sOutName, int iMaxOut)
{
	strcpy_s(sOutName, iMaxOut, this->sName);
	return (int)strlen(this->sName);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::ChangeName(const char *sChildName)
{
	strcpy_s(this->sName, sizeof(this->sName), sChildName);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Initialize(const char *sChildName)
{
	if(sChildName)
	{
		if((int)strlen(sChildName) > 0)
		{
			strcpy_s(this->sName, sizeof(this->sName), sChildName);
		}
		else{
			memset(this->sName, 0, sizeof(this->sName));
		}
	}
	else{
		memset(this->sName, 0, sizeof(this->sName));
	}

	this->pBuilder = new CStringBuilder;

	return true;
}
///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Initialize(void)
{
	return this->Initialize(NULL);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Destroy(void)
{
	if(this->pBuilder) delete this->pBuilder;
	this->pBuilder = NULL;
	memset(this->sName, 0, sizeof(this->sName));
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(CXMLWriter *XML)
{
	int iAlloc = XML->Length() + 1;
	int iLength = 0;

	#ifdef _USE_GLOBAL_MEMPOOL
		char *sBuf = (char *) pMem->Allocate(iAlloc, sizeof(char));
	#else
		char *sBuf = (char *) calloc(iAlloc, sizeof(char));
	#endif

	if(sBuf == NULL)
	{
		return false;
	}

	if((iLength = XML->Text(sBuf, iAlloc)) > 0)
	{
		bool bResult = this->pBuilder->Append(sBuf, iLength);

		#ifdef _USE_GLOBAL_MEMPOOL
			pMem->Free(sBuf);
		#else
			free(sBuf);
		#endif

		return bResult;
	}

	#ifdef _USE_GLOBAL_MEMPOOL
		pMem->Free(sBuf);
	#else
		free(sBuf);
	#endif

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(CXMLReader *XML)
{
	return this->pBuilder->Append(XML->Text(), XML->Length());
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, CXMLWriter *XML)
{
	return this->Add(sAttributeName, XML->pBuilder->Buffer, XML->pBuilder->Length);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, CXMLReader *XML)
{
	return this->Add(sAttributeName, XML->Text(), XML->Length());
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, const char *sValue)
{
	return this->Add(sAttributeName, sValue, (int)strlen(sValue));
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, const char *sValue, const int iValueSz)
{
	char sTag[255];

	sprintf_s(sTag, sizeof(sTag), "<%s>", sAttributeName);
	this->pBuilder->Append(sTag);

	this->pBuilder->Append(sValue, iValueSz);

	sprintf_s(sTag, sizeof(sTag), "</%s>", sAttributeName);
	this->pBuilder->Append(sTag);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, const double dValue, const int iDecimalPlaces)
{
	char sFormat[16];
	sprintf_s(sFormat, sizeof(sFormat), "<%%s>%%.%df</%%s>", iDecimalPlaces);
	this->pBuilder->sPrintF(sFormat, sAttributeName, dValue, sAttributeName);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, const double dValue)
{
	this->pBuilder->sPrintF("<%s>%f</%s>", sAttributeName, dValue, sAttributeName);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, const signed int iValue)
{
	this->pBuilder->sPrintF("<%s>%d</%s>", sAttributeName, iValue, sAttributeName);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, const DWORD dwValue)
{
	this->pBuilder->sPrintF("<%s>%d</%s>", sAttributeName, dwValue, sAttributeName);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, const unsigned int iValue)
{
	this->pBuilder->sPrintF("<%s>%d</%s>", sAttributeName, iValue, sAttributeName);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

bool CXMLWriter::Add(const char *sAttributeName, const __int64 i64Value)
{
	this->pBuilder->sPrintF("<%s>%I64d</%s>", sAttributeName, i64Value, sAttributeName);
	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	Could not overload due to conflicts with other functions.
*/
bool CXMLWriter::AddBool(const char *sAttributeName, const bool bValue)
{
	return this->Add(sAttributeName, bValue ? "Yes" : "No");
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	The returned pointer must be freed externally.
*/
#ifdef _CXML_USE_COMPRESSION
char *CXMLWriter::Compress(int *iOutLength)
{
	return this->Compress(iOutLength, 9);
}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	The returned pointer must be freed externally.
*/
#ifdef _CXML_USE_COMPRESSION
char *CXMLWriter::Compress(int *iOutLength, int iCompressionLevel)
{
	return CompressXML(this, iOutLength, iCompressionLevel);
}
#endif

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	The returned pointer must be freed externally.
*/
char *CXMLWriter::Text(void)
{
	int iOutSz = 0;
	return this->Text(&iOutSz);
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	The returned pointer must be freed externally.
*/

char *CXMLWriter::Text(int *iOutSz)
{
	int iMaxSize = this->Length() + 1;
	if(iMaxSize > 0)
	{
		char *sXML = (char *) calloc(iMaxSize, sizeof(char));
		if(sXML)
		{
			*iOutSz = this->Text(sXML, iMaxSize);
			return sXML;
		}
	}

	return NULL;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

int CXMLWriter::Text(char *sOutXML, int iMaxOutSz)
{
	this->pBuilder->Terminate();

	char sTag[255];
	bool bHasName = ((int)strlen(this->sName) > 0);

	if(bHasName)
	{
		sprintf_s(sTag, sizeof(sTag), "<%s>", this->sName);
		strcpy_s(sOutXML, iMaxOutSz, sTag);
	}

	strcat_s(sOutXML, iMaxOutSz, this->pBuilder->Buffer);

	if(bHasName)
	{
		sprintf_s(sTag, sizeof(sTag), "</%s>", this->sName);
		strcat_s(sOutXML, iMaxOutSz, sTag);
	}

	return this->Length() - 1; //Return the actual size, excluding the null terminator.
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

//The minimum ammount of RAM required to store the XML (including the NULL terminator).
int CXMLWriter::Length(void)
{
	int iTotalSize = 0;

	if((iTotalSize = (int)strlen(this->sName)) > 0)
	{
		iTotalSize *= 2; //One name in the opening tag, and on in the closing tag.
		iTotalSize += 6; // Extra space for "<>", "</>" and the NULL terminator.
	}

	return iTotalSize + this->pBuilder->Length;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

CStringBuilder *CXMLWriter::GetBuilder(void)
{
	return this->pBuilder;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/*
	//Writes XML to a file.
*/
bool CXMLWriter::ToFile(const char *sFileName)
{
	HANDLE hTarget = NULL;

	hTarget = CreateFile(sFileName, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hTarget == INVALID_HANDLE_VALUE)
	{
		return false; //File not found or access denied.
	}

	DWORD dwBytesWritten = 0;

	int iNameLen = (int)strlen(this->sName);

	if(iNameLen > 0)
	{
		WriteFile(hTarget, "<", 1, &dwBytesWritten, NULL);
		WriteFile(hTarget, this->sName, iNameLen, &dwBytesWritten, NULL);
		WriteFile(hTarget, ">", 1, &dwBytesWritten, NULL);
	}

	WriteFile(hTarget, this->pBuilder->Buffer, this->pBuilder->Length, &dwBytesWritten, NULL);

	if(iNameLen > 0)
	{
		WriteFile(hTarget, "</", 2, &dwBytesWritten, NULL);
		WriteFile(hTarget, this->sName, iNameLen, &dwBytesWritten, NULL);
		WriteFile(hTarget, ">", 1, &dwBytesWritten, NULL);
	}

	CloseHandle(hTarget);

	return true;
}

///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
#endif
